@@ -33,7 +33,7 @@
 #if defined(STDC_HEADERS)
 #include <stdlib.h>
 #endif
-#if defined(USG) || defined(STDC_HEADERS)
+#if HAVE_STRING_H || defined(STDC_HEADERS)
 #include <string.h>
 #if !defined(STDC_HEADERS)
 #include <memory.h>
@@ -67,7 +67,7 @@ bcopy (from, to, len)
 #endif
 #endif
 
-char *version_string = "GNU sed version 1.14b";
+char *version_string = "GNU sed version 1.15";
 
 /* Struct vector is used to describe a chunk of a compiled sed program.  
  * There is one vector for the main program, and one for each { } pair,
@@ -346,6 +346,7 @@ static char NO_ADDR[] = "Command doesn't take any addresses";
 static char LINE_JUNK[] = "Extra characters after command";
 static char BAD_EOF[] = "Unexpected End-of-file";
 static char NO_REGEX[] = "No previous regular expression";
+static char NO_COMMAND[] = "Missing command";
 
 static struct option longopts[] =
 {
@@ -400,6 +401,7 @@ main (argc, argv)
 	  break;
 	case 'V':
 	  fprintf (stderr, "%s\n", version_string);
+	  exit (0);
 	  break;
 	default:
 	  usage ();
@@ -545,7 +547,11 @@ compile_program (vector, open_line)
     {
     skip_comment:
       do
-	ch = inchar ();
+	{
+	  ch = inchar ();
+	  if (ch == '\\')
+	    ch = inchar ();
+	}
       while (ch != EOF && (isblank (ch) || ch == '\n' || ch == ';'));
       if (ch == EOF)
 	break;
@@ -553,7 +559,10 @@ compile_program (vector, open_line)
 
       if (vector->v_length == vector->v_allocated)
 	{
-	  vector->v = (struct sed_cmd *) ck_realloc ((VOID *) vector->v, (vector->v_length + MORE_CMDS) * sizeof (struct sed_cmd));
+	  vector->v = ((struct sed_cmd *)
+		       ck_realloc ((VOID *) vector->v,
+				   ((vector->v_length + MORE_CMDS)
+				    * sizeof (struct sed_cmd))));
 	  vector->v_allocated += MORE_CMDS;
 	}
       cur_cmd = vector->v + vector->v_length;
@@ -584,11 +593,11 @@ compile_program (vector, open_line)
       if (cur_cmd->a1.addr_type == addr_is_num
 	  && cur_cmd->a2.addr_type == addr_is_num
 	  && cur_cmd->a2.addr_number < cur_cmd->a1.addr_number)
-	cur_cmd->a2.addr_number = cur_cmd->a1.addr_number + 1;
+	cur_cmd->a2.addr_number = cur_cmd->a1.addr_number;
 
       ch = inchar ();
       if (ch == EOF)
-	break;
+	bad_prog (NO_COMMAND);
     new_cmd:
       switch (ch)
 	{
@@ -608,7 +617,7 @@ compile_program (vector, open_line)
 	    ch = inchar ();
 	  while (ch != EOF && isblank (ch));
 	  if (ch == EOF)
-	    bad_prog (BAD_EOF);
+	    bad_prog (NO_COMMAND);
 #if 0
 	  savchar (ch);
 #endif
@@ -736,7 +745,7 @@ compile_program (vector, open_line)
 	    }
 	  if (ch != slash)
 	    {
-	      if (ch == '\n')
+	      if (ch == '\n' && prog_line > 1)
 		--prog_line;
 	      bad_prog ("Unterminated `s' command");
 	    }
@@ -851,7 +860,7 @@ void
 bad_prog (why)
      char *why;
 {
-  if (prog_line)
+  if (prog_line > 0)
     fprintf (stderr, "%s: file %s line %d: %s\n",
 	     myname, prog_name, prog_line, why);
   else
@@ -965,10 +974,10 @@ compile_regex (slash)
 {
   VOID *b;
   int ch;
-  int in_char_class = 0;
+  int char_class_pos = -1;
 
   b = init_buffer ();
-  while ((ch = inchar ()) != EOF && (ch != slash || in_char_class))
+  while ((ch = inchar ()) != EOF && (ch != slash || (char_class_pos >= 0)))
     {
       if (ch == '^')
 	{
@@ -996,8 +1005,9 @@ compile_regex (slash)
 	}
       else if (ch == '[')
 	{
+	  if (char_class_pos < 0)
+	    char_class_pos = size_buffer (b);
 	  add1_buffer (b, ch);
-	  in_char_class = 1;
 	  continue;
 	}
       else if (ch == ']')
@@ -1005,20 +1015,16 @@ compile_regex (slash)
 	  add1_buffer (b, ch);
 	  {
 	    char * regexp = get_buffer (b);
-	    int len = size_buffer (b);
-	    /* In regexp syntax, the close first brace doesn't conclude
-	     * either character set: "[]xyz]" or "[^]abc]".  We know
-	     * we've read at least "[" and "]" so len >= 2.
-	     */
-	    if (   (regexp[len - 2] != '[')
-		&& (   (len == 2)
-		    || (regexp[len - 2] != '^')
-		    || (regexp[len - 3] != ']')))
-	      in_char_class = 0;
+	    int pos = size_buffer (b) - 1;
+	    if (!(   (char_class_pos >= 0)
+		  && (   (pos == char_class_pos + 1)
+		      || (   (pos == char_class_pos + 2)
+			  && (regexp[char_class_pos + 1] == '^')))))
+	      char_class_pos = -1;
 	    continue;
 	  }
 	}
-      else if (ch != '\\' || in_char_class)
+      else if (ch != '\\' || (char_class_pos >= 0))
 	{
 	  add1_buffer (b, ch);
 	  continue;
@@ -1194,14 +1200,14 @@ read_file (name)
 }
 
 static char *
-chr_pos (str, chr, len)
+eol_pos (str, len)
      char *str;
-     char chr;
      int len;
 {
   while (len--)
-    if (*str++ == chr)
+    if (*str++ == '\n')
       return --str;
+  return --str;
 }
 
 static void
@@ -1259,7 +1265,8 @@ restart:
 	{
 	  addr_matched = 1;
 	  if (cur_cmd->a2.addr_type != addr_is_null)
-	    if (!match_address (&(cur_cmd->a2)))
+	    if (   (cur_cmd->a2.addr_type == addr_is_regex)
+		|| !match_address (&(cur_cmd->a2)))
 	      cur_cmd->aflags |= A1_MATCHED_BIT;
 
 	}
@@ -1334,7 +1341,7 @@ restart:
 	    char *tmp;
 	    int newlength;
 
-	    tmp = chr_pos (line.text, '\n', line.length);
+	    tmp = eol_pos (line.text, line.length);
 	    newlength = line.length - (tmp - line.text) - 1;
 	    if (newlength)
 	      {
@@ -1469,7 +1476,7 @@ restart:
 	  {
 	    char *tmp;
 
-	    tmp = chr_pos (line.text, '\n', line.length);
+	    tmp = eol_pos (line.text, line.length);
 	    ck_fwrite (line.text, 1,
 		       tmp ? tmp - line.text + 1
 		       : line.length, stdout);
@@ -1723,8 +1730,8 @@ read_pattern_space ()
 	{
 	  if (n == line.alloc)
 	    return 0;
-	  *p++ = '\n';
-	  --n;
+	  /* *p++ = '\n'; */
+	  /* --n; */
 	  line.length = line.alloc - n;
 	  if (last_input_file)
 	    input_EOF++;
@@ -1767,8 +1774,8 @@ append_pattern_space ()
 	{
 	  if (n == line.alloc)
 	    return;
-	  *p++ = '\n';
-	  --n;
+	  /* *p++ = '\n'; */
+	  /* --n; */
 	  line.length = line.alloc - n;
 	  if (last_input_file)
 	    input_EOF++;
